import unittest
import io
import json

import riu.journal
import riu.time


class Test(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        self.maxDiff = None
        super().__init__(*args, **kwargs)

    def test_journalize(self):

        # Add entries

        s = io.StringIO()

        timestamp_dt1 = riu.time.utcnow()

        riu.journal.journalize(
            s,
            'type1',
            timestamp_dt=timestamp_dt1,
            aa='bb',
            cc='dd')

        timestamp_dt2 = riu.time.utcnow()

        riu.journal.journalize(
            s,
            'type1',
            timestamp_dt=timestamp_dt2,
            ee='ff',
            gg='hh')

        timestamp_dt3 = riu.time.utcnow()

        riu.journal.journalize(
            s,
            'type2',
            timestamp_dt=timestamp_dt3,
            ii='jj',
            kk='ll')


        # Check

        s.seek(0)

        records = [
            json.loads(line)
            for line
            in s
        ]

        expected = [
            {
                'aa': 'bb',
                'cc': 'dd',
                't': 'type1',
                't2': timestamp_dt1.isoformat(),
            },
            {
                'ee': 'ff',
                'gg': 'hh',
                't': 'type1',
                't2': timestamp_dt2.isoformat(),
            },
            {
                'ii': 'jj',
                'kk': 'll',
                't': 'type2',
                't2': timestamp_dt3.isoformat(),
            }
        ]

        self.assertEqual(records, expected)


        # Ingest using utility

        records = riu.journal.parse_journal_stream_gen(s)
        records = list(records)

        self.assertEqual(records, expected)


        # Ingest using utility (with filter)

        expected = [
            {
                'ii': 'jj',
                'kk': 'll',
                't': 'type2',
                't2': timestamp_dt3.isoformat(),
            }
        ]

        records = riu.journal.parse_journal_stream_gen(s, filter_types=['type2'])
        records = list(records)

        self.assertEqual(records, expected)


    test_parse_journal_stream = test_journalize



import datetime

import dateutil.tz


def utcnow(round=False):

    now_dt = \
        datetime.datetime.utcnow().replace(
            tzinfo=dateutil.tz.UTC)

    if round is True:
        now_dt -= datetime.timedelta(microseconds=now_dt.microsecond)

    return now_dt



import os
import logging

_LOGGER = logging.getLogger(__name__)


def get_module_from_module_filepath(module_filepath):

    with open(module_filepath) as f:
        source = f.read()

    c = compile(source, module_filepath, 'exec')

    g = globals()
    g = g.copy()

    exec(c, g)

    return g


def get_module_symbol_with_filepath(module_filepath, name):

    g = get_module_from_module_filepath(module_filepath)
    s = g[name]

    return s


def get_module_symbol_with_reference(root_module_path, reference):

    package_and_module, class_name = reference.rsplit('.', 1)

    rel_package_and_module_filepath_without_extension = package_and_module.replace('.', os.sep)

    package_and_module_filepath_without_extension = os.path.join(root_module_path, rel_package_and_module_filepath_without_extension)
    package_and_module_filepath = package_and_module_filepath_without_extension + '.py'

    _LOGGER.debug("Translating reference [{}]: "
                  "([{}] + [{}]) -> [{}] -> [{}] [{}]".format(
                  reference, root_module_path,
                  rel_package_and_module_filepath_without_extension,
                  package_and_module_filepath_without_extension,
                  package_and_module_filepath, class_name))

    s = get_module_symbol_with_filepath(
            package_and_module_filepath,
            class_name)

    return s



import logging
import json

import riu.time

_LOGGER = logging.getLogger(__name__)


def journalize(resource, type_=None, timestamp_dt=None, **kwargs):

    if timestamp_dt is None:
        timestamp_dt = riu.time.utcnow()

    info = {
        't2': timestamp_dt.isoformat(),
    }

    if type_ is not None:
        info['t'] = type_

    info.update(kwargs)

    json.dump(info, resource)
    resource.write('\n')


def parse_journal_stream_gen(s, filter_types=None):

    s.seek(0)

    journal_entries = (
        json.loads(line)
        for line
        in s
    )

    if filter_types:
        journal_entries = (
            record
            for record
            in journal_entries
            if record['t'] in filter_types
        )

    yield from journal_entries



import os
import logging
import json
import traceback
import contextlib
import shutil
import tempfile

_LOGGER = logging.getLogger(__name__)


def get_pretty_json(data):

# TODO(dustin): Add test

    return \
        json.dumps(
            data,
            sort_keys=True,
            indent=4,
            separators=(',', ': '))


def write_pretty_json(data, f):

# TODO(dustin): Add test

    json.dump(
        data,
        f,
        sort_keys=True,
        indent=4,
        separators=(',', ': '))

    f.write('\n')


def format_exception(e):

# TODO(dustin): Add test

    lines = traceback.format_exception(e.__class__, e, e.__traceback__)
    return ''.join(lines)


@contextlib.contextmanager
def temp_path():

# TODO(dustin): Add test

    original_wd = os.getcwd()

    path = tempfile.mkdtemp()
    os.chdir(path)

    try:
        yield path
    finally:
        os.chdir(original_wd)

        try:
            shutil.rmtree(path)
        except:
            pass



import os
import unittest
import contextlib

import riu.plugin
import riu.utility


_TEST_MODULE = """\
class TestModule(object):
    def test_method(self):
        return 123
"""


class Test(unittest.TestCase):
    @contextlib.contextmanager
    def _stage_gen(self):
        with riu.utility.temp_path() as temp_path:
            filename = 'test_module.py'
            filepath = os.path.join(temp_path, filename)

            with open(filepath, 'w') as f:

                # Write

                f.write(_TEST_MODULE)
                f.flush()

            yield temp_path, filepath

    def test_get_module_from_module_filepath(self):
        with self._stage_gen() as (temp_path, filepath):

            # Retrieve

            g = riu.plugin.get_module_from_module_filepath(filepath)


        # Instantiate

        cls = g['TestModule']
        o = cls()


        # Call method

        actual = o.test_method()
        self.assertEqual(actual, 123)

    def test_get_module_symbol_with_filepath(self):
        with self._stage_gen() as (temp_path, filepath):

            # Retrieve

            cls = riu.plugin.get_module_symbol_with_filepath(
                    filepath,
                    'TestModule')


        # Instantiate and call method

        o = cls()

        actual = o.test_method()
        self.assertEqual(actual, 123)

    def test_get_module_symbol_with_reference(self):
        with self._stage_gen() as (temp_path, filepath):

            cls = riu.plugin.get_module_symbol_with_reference(
                    temp_path,
                    'test_module.TestModule')


        # Instantiate and call method

        o = cls()

        actual = o.test_method()
        self.assertEqual(actual, 123)



import logging

_LOGGER = logging.getLogger(__name__)


class BaseEntity(object):
    def __init__(self, data):
        self._data = data

    @property
    def data(self):
        return self._data

    @property
    def id(self):
        raise NotImplementedError()

    @property
    def id_attribute_name(self):
        return 'id'



import unittest

import riu.hierarchy


_COMPLEX_MAPPINGS = {
    'a': 'AA',
    'c': 'CC',
    'e': 'EE',
    ('e', 'f'): 'FF',
    ('e', 'h'): 'HH',
    ('e', 'j'): 'JJ',
    ('e', 'j', 'k'): 'KK',
    ('e', 'j', 'm'): 'MM',
}


class _TestNode(riu.hierarchy.BaseTreeNode):
    def __init__(self, data):
        self._data = data

    @property
    def id(self):
        return self._data['id']

    @property
    def children(self):
        return self._data['children']

    @property
    def raw(self):

        exported = {
            'id': self.id,
            'children': [
                child.raw
                for child
                in self.children
            ]
        }

        return exported


def _TEST_NODE_FACTORY(record, children):
    record = record.copy()
    record['children'] = children

    return _TestNode(record)


class Test(unittest.TestCase):
    def test_translate_hierarchy_with_mappings(self):

        data = {
            'a': 'b',
            'c': 'd',
        }

        mappings = {
            'a': 'AA',
            'c': 'CC',
        }

        translated = \
            riu.hierarchy.translate_hierarchy_with_mappings(
                data,
                mappings)

        expected = {
            'AA': 'b',
            'CC': 'd',
        }

        self.assertEqual(translated, expected)


        # Translate with an uncovered attribute (should fail)

        data['e'] = 'f'

        try:
            riu.hierarchy.translate_hierarchy_with_mappings(
                data,
                mappings)

        except KeyError:
            pass

        else:
            raise Exception("Expected KeyError for uncovered attribute.")


        # Translate with an uncovered attribute (not requiring coverage)

        translated = \
            riu.hierarchy.translate_hierarchy_with_mappings(
                data,
                mappings,
                do_require_all=False)

        expected = {
            'AA': 'b',
            'CC': 'd',

            # Just copied verbatim
            'e': 'f',
        }

        self.assertEqual(translated, expected)

    def test_translate_hierarchy_with_mappings__nested(self):

        data4 = [
            {
                'n': 'o',
                'p': 'q',
            },
            {
                'r': 's',
                't': 'u',
            },
        ]

        data3 = {
            'k': 'l',
            'm': data4,
        }

        data2 = {
            'f': 'g',
            'h': 'i',
            'j': data3,
        }

        data1 = {
            'a': 'b',
            'c': 'd',
            'e': data2,
        }

        mappings = {
            'a': 'AA',
            'c': 'CC',
            'e': 'EE',
            ('e', 'f'): 'FF',
            ('e', 'h'): 'HH',
            ('e', 'j'): 'JJ',
            ('e', 'j', 'k'): 'KK',
            ('e', 'j', 'm'): 'MM',

            # These are wrapped into a list, in the source (outer) data
            ('e', 'j', 'm', 'n'): 'NN',
            ('e', 'j', 'm', 'p'): 'PP',
            ('e', 'j', 'm', 'r'): 'RR',
            ('e', 'j', 'm', 't'): 'TT',
        }

        translated = \
            riu.hierarchy.translate_hierarchy_with_mappings(
                data1,
                mappings)

        expected = {
            'AA': 'b',
            'CC': 'd',
            'EE': {
                'FF': 'g',
                'HH': 'i',
                'JJ': {
                    'KK': 'l',
                    'MM': [
                        {
                            'NN': 'o',
                            'PP': 'q',
                        },
                        {
                            'RR': 's',
                            'TT': 'u',
                        },
                    ],
                },
            },
        }

        self.assertEqual(translated, expected)


        # Translate with an uncovered attribute (should fail)

        data3['v'] = 'w'

        try:
            riu.hierarchy.translate_hierarchy_with_mappings(
                data1,
                mappings)

        except KeyError:
            pass

        else:
            raise Exception("Expected KeyError for uncovered attribute.")


        # Translate with an uncovered attribute (not requiring coverage)

        translated = \
            riu.hierarchy.translate_hierarchy_with_mappings(
                data1,
                mappings,
                do_require_all=False)

        expected = {
            'AA': 'b',
            'CC': 'd',
            'EE': {
                'FF': 'g',
                'HH': 'i',
                'JJ': {
                    # Just copied verbatim
                    'v': 'w',

                    'KK': 'l',
                    'MM': [
                        {
                            'NN': 'o',
                            'PP': 'q',
                        },
                        {
                            'RR': 's',
                            'TT': 'u',
                        },
                    ],
                },
            },
        }

        self.assertEqual(translated, expected)

    def test_invert_mappings_tuple(self):

        # Check a depth of 1

        cache = {}

        self.assertEqual(
            riu.hierarchy._invert_mappings_tuple(_COMPLEX_MAPPINGS, ('a',), cache),
            ('AA',))


        expected = {}
        self.assertEqual(cache, expected)


        # Check a depth of 2

        cache = {}

        self.assertEqual(
            riu.hierarchy._invert_mappings_tuple(_COMPLEX_MAPPINGS, ('e', 'f'), cache),
            ('EE', 'FF'))

        expected = {
            ('e', 'f'): ('EE', 'FF')
        }

        self.assertEqual(cache, expected)


        # Check a depth of 3

        cache = {}

        self.assertEqual(
            riu.hierarchy._invert_mappings_tuple(_COMPLEX_MAPPINGS, ('e', 'j', 'k'), cache),
            ('EE', 'JJ', 'KK'))

        expected = {
            ('e', 'j'): ('EE', 'JJ'),
            ('e', 'j', 'k'): ('EE', 'JJ', 'KK'),
        }

        self.assertEqual(cache, expected)

    def test_get_reverse_hierarchy_mappings(self):

        rmappings = riu.hierarchy.get_reverse_hierarchy_mappings(_COMPLEX_MAPPINGS)

        expected = {
            'AA': 'a',
            'CC': 'c',
            'EE': 'e',
            ('EE', 'FF'): 'f',
            ('EE', 'HH'): 'h',
            ('EE', 'JJ'): 'j',
            ('EE', 'JJ', 'KK'): 'k',
            ('EE', 'JJ', 'MM'): 'm',
        }

        self.assertEqual(rmappings, expected)


        # If we reverse them again, we should end-up with the original mappings

        recovered_mappings = \
            riu.hierarchy.get_reverse_hierarchy_mappings(
                rmappings)

        self.assertEqual(recovered_mappings, _COMPLEX_MAPPINGS)

    def test_pick_from_hierarchical_records_gen(self):

        records = [
            {"timestamp_phrase": "2024-03-06T13:28:59.573663+00:00", "event_name": "create_variant", "metadata": {"write": True, "config": ["_managed:used and excellent", "large:", "grey/black/red logo:#868484"], "product_id": 8200540651811, "variant_id": 47692705431843}},
            {"timestamp_phrase": "2024-03-06T13:28:59.573721+00:00", "event_name": "want_to_upload_image_primary", "metadata": {"product_id": 8200540651811, "variant_id": 47692705431843, "config": ["_managed:used and excellent", "large:", "grey/black/red logo:#868484"], "image_filepath": "/output/workflow/dropbox/images/2024-01-10/bld00000000000150400.jpg"}},
            {"timestamp_phrase": "2024-03-06T13:29:00.629621+00:00", "event_name": "create_variant", "metadata": {"write": True, "config": ["_managed:used and excellent", "large:", "grey/black:#000000"], "product_id": 8200540651811, "variant_id": 47692706021667}},
            {"timestamp_phrase": "2024-03-06T13:29:00.629771+00:00", "event_name": "want_to_upload_image_primary", "metadata": {"product_id": 8200540651811, "variant_id": 47692706021667, "config": ["_managed:used and excellent", "large:", "grey/black:#000000"], "image_filepath": "/output/workflow/dropbox/images/2024-01-10/bld00000000000123957.jpg"}},
            {"timestamp_phrase": "2024-03-06T13:29:01.724781+00:00", "event_name": "create_variant", "metadata": {"write": True, "config": ["_managed:used and good", "2X large:", "black:#000000"], "product_id": 8200540651811, "variant_id": 47692706087203}},
            {"timestamp_phrase": "2024-03-06T13:29:01.725432+00:00", "event_name": "want_to_upload_image_primary", "metadata": {"product_id": 8200540651811, "variant_id": 47692706087203, "config": ["_managed:used and good", "2X large:", "black:#000000"], "image_filepath": "/output/workflow/dropbox/images/2024-01-10/bld00000000000123921.jpg"}},
            {"timestamp_phrase": "2024-03-06T13:29:02.466462+00:00", "event_name": "variant_already_has_images", "metadata": {"product_id": 8200540651811, "config": ["_managed:used and excellent", "extra large:", "black:#000000"], "variant_id": 47356921250083}},
            {"timestamp_phrase": "2024-03-06T13:29:02.466503+00:00", "event_name": "variant_already_has_images", "metadata": {"product_id": 8200540651811, "config": ["_managed:used and excellent", "medium:", "black:#000000"], "variant_id": 47124530463011}},
            {"timestamp_phrase": "2024-03-06T13:29:02.466529+00:00", "event_name": "variant_already_has_images", "metadata": {"product_id": 8200540651811, "config": ["_managed:used and good", "2x large:", "black:#000000"], "variant_id": 47124530495779}},
            {"timestamp_phrase": "2024-03-06T13:29:02.466550+00:00", "event_name": "variant_already_has_images", "metadata": {"product_id": 8200540651811, "config": ["_managed:used and good", "large:", "black:#000000"], "variant_id": 47124530528547}},
        ]

        mappings2 = {
            'timestamp_phrase': 'final_timestamp_phrase',
            'metadata.product_id': 'final_product_id',
        }

        g = riu.hierarchy.pick_from_hierarchical_records_gen(records, mappings2)

        picked = list(g)

        expected = [
            { 'final_timestamp_phrase': '2024-03-06T13:28:59.573663+00:00', 'final_product_id': 8200540651811 },
            { 'final_timestamp_phrase': '2024-03-06T13:28:59.573721+00:00', 'final_product_id': 8200540651811 },
            { 'final_timestamp_phrase': '2024-03-06T13:29:00.629621+00:00', 'final_product_id': 8200540651811 },
            { 'final_timestamp_phrase': '2024-03-06T13:29:00.629771+00:00', 'final_product_id': 8200540651811 },
            { 'final_timestamp_phrase': '2024-03-06T13:29:01.724781+00:00', 'final_product_id': 8200540651811 },
            { 'final_timestamp_phrase': '2024-03-06T13:29:01.725432+00:00', 'final_product_id': 8200540651811 },
            { 'final_timestamp_phrase': '2024-03-06T13:29:02.466462+00:00', 'final_product_id': 8200540651811 },
            { 'final_timestamp_phrase': '2024-03-06T13:29:02.466503+00:00', 'final_product_id': 8200540651811 },
            { 'final_timestamp_phrase': '2024-03-06T13:29:02.466529+00:00', 'final_product_id': 8200540651811 },
            { 'final_timestamp_phrase': '2024-03-06T13:29:02.466550+00:00', 'final_product_id': 8200540651811 },
        ]

        self.assertEqual(
            sorted(picked, key=lambda record: record['final_timestamp_phrase']),
            sorted(expected, key=lambda record: record['final_timestamp_phrase']))

    def test_build_tree_node__dict(self):

        # Test with single node

        record1 = {
        }

        record_index = {
            1: record1,
        }

        out_node_index = {}
        root_id = 1
        child_attribute = 'children'

        node = riu.hierarchy._build_tree_node(
                record_index,
                out_node_index,
                child_attribute,
                root_id,
                node_factory_fn=None)

        expected_node = {'children': []}

        self.assertEqual(node, expected_node)

        expected = {
            1: expected_node
        }

        self.assertEqual(out_node_index, expected)


        # Test with multiple nodes

        record1 = {
            child_attribute: [
                2,
            ],
        }

        record2 = {
        }

        record_index = {
            1: record1,
            2: record2,
        }

        out_node_index = {}
        root_id = 1

        node = riu.hierarchy._build_tree_node(
                record_index,
                out_node_index,
                child_attribute,
                root_id,
                node_factory_fn=None)

        expected_node2 = {'children': []}
        expected_node1 = {'children': [expected_node2]}

        self.assertEqual(node, expected_node1)

        expected = {
            1: expected_node1,
            2: expected_node2,
        }

        self.assertEqual(out_node_index, expected)

    def test_build_tree_node__node_class(self):

        # Test with single node

        record1 = {
            'id': 1,
        }

        record_index = {
            1: record1,
        }

        out_node_index = {}
        root_id = 1
        child_attribute = 'children'

        node = riu.hierarchy._build_tree_node(
                record_index,
                out_node_index,
                child_attribute,
                root_id,
                node_factory_fn=_TEST_NODE_FACTORY)

        expected_node1_raw = {
            'id': 1,
            'children': [],
        }

        self.assertEqual(node.raw, expected_node1_raw)

        out_node_index_flat = {
            id_: node.raw
            for id_, node
            in out_node_index.items()
        }

        expected = {
            1: expected_node1_raw
        }

        self.assertEqual(out_node_index_flat, expected)


        # Test with multiple nodes

        record1 = {
            'id': 1,

            child_attribute: [
                2,
            ],
        }

        record2 = {
            'id': 2,
        }

        record_index = {
            1: record1,
            2: record2,
        }

        out_node_index = {}
        root_id = 1

        node = riu.hierarchy._build_tree_node(
                record_index,
                out_node_index,
                child_attribute,
                root_id,
                node_factory_fn=_TEST_NODE_FACTORY)

        expected_node2_raw = {'id': 2, 'children': []}
        expected_node1_raw = {'id': 1, 'children': [expected_node2_raw]}

        self.assertEqual(node.raw, expected_node1_raw)

        out_node_index_flat = {
            id_: node.raw
            for id_, node
            in out_node_index.items()
        }

        expected = {
            1: expected_node1_raw,
            2: expected_node2_raw,
        }

        self.assertEqual(out_node_index_flat, expected)

    def test_tree_build_down__dict(self):

        records = [
            {
                'id': 1,
                'children': [
                    2,
                    3,
                ],
            },

            {
                'id': 2,
                'children': [

                ],
            },

            {
                'id': 3,
                'children': [
                    4,
                ],
            },

            {
                'id': 4,
                'children': [

                ],
            },
        ]

        id_attribute = 'id'
        child_attribute = 'children'

        node_index, \
        root_id, \
        leaf_ids = \
            riu.hierarchy.tree_build_down(
                records,
                id_attribute,
                child_attribute,
                node_factory_fn=None)

        self.assertEqual(root_id, 1)
        self.assertEqual(leaf_ids, [2, 4])


        node2 = {
            'id': 2,
            'children': [

            ],
        }

        node4 = {
            'id': 4,
            'children': [

            ],
        }

        node3 = {
            'id': 3,
            'children': [
                node4,
            ],
        }

        node1 = {
            'id': 1,
            'children': [
                node2,
                node3,
            ],
        }

        expected = {
            1: node1,
            2: node2,
            3: node3,
            4: node4,
        }

        self.assertEqual(node_index, expected)

    def test_tree_build_down__node_class(self):

        records = [
            {
                'id': 1,
                'children': [
                    2,
                    3,
                ],
            },

            {
                'id': 2,
                'children': [

                ],
            },

            {
                'id': 3,
                'children': [
                    4,
                ],
            },

            {
                'id': 4,
                'children': [

                ],
            },
        ]

        id_attribute = 'id'
        child_attribute = 'children'

        out_node_index, \
        root_id, \
        leaf_ids = \
            riu.hierarchy.tree_build_down(
                records,
                id_attribute,
                child_attribute,
                node_factory_fn=_TEST_NODE_FACTORY)

        self.assertEqual(root_id, 1)
        self.assertEqual(leaf_ids, [2, 4])


        node2 = _TestNode({
            'id': 2,
            'children': [

            ],
        })

        node4 = _TestNode({
            'id': 4,
            'children': [

            ],
        })

        node3 = _TestNode({
            'id': 3,
            'children': [
                node4,
            ],
        })

        node1 = _TestNode({
            'id': 1,
            'children': [
                node2,
                node3,
            ],
        })

        out_node_index_flat = {
            id_: node.raw
            for id_, node
            in out_node_index.items()
        }

        expected = {
            1: node1.raw,
            2: node2.raw,
            3: node3.raw,
            4: node4.raw,
        }

        self.assertEqual(out_node_index_flat, expected)

    def test_scalar_tree_build_down(self):

        class Node(riu.hierarchy.BaseTreeNode):
            def __init__(self, id_, name, children=None):
                if children is None:
                    children = []

                self._id = id_
                self._children = children

                self.name = name

                super().__init__({})

            @property
            def id(self):
                return self._id

            @property
            def children(self):
                return self._children


        node5 = Node(55, 'ee')
        node4 = Node(44, 'dd', children=[node5])
        node3 = Node(33, 'cc', children=[node4])
        node2 = Node(22, 'bb')
        node1 = Node(11, 'aa', children=[node2, node3])

        name_extractor_fn = lambda node: node.name

        flat_root_node = \
            riu.hierarchy.scalar_tree_build_down(
                node1,
                name_extractor_fn)

        expected = {
            'aa': {
                'bb': {
                    None: 22
                },

                'cc': {
                    'dd': {
                        'ee': {
                            None: 55
                        },
                        None: 44
                    },
                    None: 33
                },
                None: 11
            }
        }

        self.assertEqual(flat_root_node, expected)



import os
import setuptools

with open(os.path.join('riu', 'resources', 'README.md')) as f:
    long_description = f.read()

_VERSION = '0.0.4'

_REQUIREMENTS = [
    'build',
    'twine',
    'python-dateutil',
]

setuptools.setup(
    name='ri-common',
    version=_VERSION,
    description="Common, reusable utilities",
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[],
    author='Dustin Oprea',
    author_email='dustin@randomingenuity.com',
    packages=setuptools.find_packages(exclude=['tests']),
    include_package_data=True,
    zip_safe=False,
    package_data={
        'riu': [
            'resources/README.md',
            'resources/requirements.txt',
            'resources/requirements-testing.txt',
        ],
    },
    install_requires=_REQUIREMENTS,
    scripts=[
    ],
)



import logging

import riu.entity

_LOGGER = logging.getLogger(__name__)


def translate_hierarchy_with_mappings(
        data, mappings, outer_parent_tuple=None, level=0, do_require_all=True,
        do_include_untranslated=True):
    """Translate names in a hierarchy according to a mappings dictionary. Deeper
    elements are represented as tuples (though the top-level ones can be as
    well). the `outer` name is the human name and the `inner` name is the
    internal/storage name [that we're trying to obscure via this functionality].

    Typically we want all attributes to be present if we're using this
    functionality at all. `do_require_all` should be set to False to support any
    preexisting data that is not covered.
    """

    # Allows us to translate lists of blobs
    if isinstance(data, list) is True:

        results = []
        for i, x in enumerate(data):

            y = translate_hierarchy_with_mappings(
                    x,
                    mappings,
                    outer_parent_tuple=outer_parent_tuple,
                    level=level + 1,
                    do_require_all=do_require_all)

            results.append(y)

        return results

    elif isinstance(data, dict) is False:

        # Some scalar (or, less likely, an unsupported type, which wouldn't
        # make sense in this context since we're writing data). Just pass it
        # through.

        return data


    # It's a dictionary

    translated = {}
    for outer_name, value in data.items():

        # This is what we're look for the translation under. It can be a string
        # or a tuple.
        if outer_parent_tuple is None:
            reference_outer_tuple = outer_name
        else:
            reference_outer_tuple = outer_parent_tuple + (outer_name,)

        try:
            inner_name = mappings[reference_outer_tuple]

        except KeyError:

            # If not enforcing then just set into the output dictionary verbatim
            if do_require_all is False:
                if do_include_untranslated is True:
                    translated[outer_name] = value

                continue

            raise


        # Descend

        # This is the parent-tuple that we'll be passing forward. It will
        # always be a tuple.
        if outer_parent_tuple is None:
            outer_parent_tuple2 = (outer_name,)
        else:
            outer_parent_tuple2 = outer_parent_tuple + (outer_name,)


        updated_value = \
            translate_hierarchy_with_mappings(
                value,
                mappings,
                outer_parent_tuple=outer_parent_tuple2,
                level=level + 1,
                do_require_all=do_require_all)

        translated[inner_name] = updated_value


    return translated


def _invert_mappings_tuple(mappings, input_key_tuple, cache):

    # If it's a 1-tuple, then try looking it up as a string. If not found, the
    # mapping is not complete and it's an error.

    if len(input_key_tuple) == 1:
        input_name = input_key_tuple[0]
        output_name = mappings[input_name]
        return (output_name,)


    # Prevent rerecursions

    try:
        return cache[input_key_tuple]
    except KeyError:
        pass


    # We have a tuple that isn't in the mapping. It must be an intermediate
    # form.

    # This retains the tuple type
    prefix_input_key_tuple = input_key_tuple[:-1]

    output_tuple = \
        _invert_mappings_tuple(
            mappings,
            prefix_input_key_tuple,
            cache)

    output_key_name = mappings[input_key_tuple]
    output_key_tuple = output_tuple + (output_key_name,)


    # Memoize so siblings caller contexts can benefit

    cache[input_key_tuple] = output_key_tuple


    return output_key_tuple


def get_reverse_hierarchy_mappings(mappings):
    """Returns a set of reverse mappings that can be provided to
    `translate_hierarchy_with_mappings` alongside already-mapped data in order
    to restore the pre-mapped data.
    """

    rmappings = {}
    cache = {}
    for outer_key, inner_member_name in mappings.items():
        if outer_key.__class__ is str:
            rmappings[inner_member_name] = outer_key

        elif outer_key.__class__ is tuple:
            inner_member_key = _invert_mappings_tuple(mappings, outer_key, cache)
            outer_member_name = outer_key[-1]

            rmappings[inner_member_key] = outer_member_name

        else:
            raise \
                Exception(
                    "We don't handle keys of type [{}]: [{}] [{}]".format(
                    key.__class__.__name__, key, value))


    return rmappings


def get_value_from_hierarchy_with_tuple_reference(record, reference):

# TODO(dustin): Add test

    parts = reference
    ptr = record
    while parts:

        # Allow us to specify nonexistent parts
        if ptr is None:
            ptr = {}

        part, parts = parts[0], parts[1:]

        try:
            if ptr.__class__ is dict:
                # This will raise KeyError if the reference is invalid
                ptr = ptr[part]
            elif ptr.__class__ is list:
                index = int(part)
                ptr = ptr[index]
            else:
                raise \
                    Exception(
                        "Can't pick [{}] from node of type [{}].".format(
                        part, ptr.__class__.__name__))

        except KeyError:
            raise

        except:
            _LOGGER.exception("Error while picking from node of type [{}]:\n"
                              "{}".format(ptr.__class__.__name__, ptr))

            raise


    return ptr


def get_value_from_hierarchy_with_string_reference(
        record, reference, separator='.'):

# TODO(dustin): Add test

    parts = reference.split(separator)
    value = get_value_from_hierarchy_with_tuple_reference(record, parts)

    return value


def pick_from_hierarchical_record(
        record, attribute_mappings, default_comma_separate_lists=False):

# TODO(dustin): Add test

    assert \
        attribute_mappings.__class__ is dict, \
        "Attribute mappings must be a dictionary: [{}]".format(
            attribute_mappings.__class__.__name__)

    assert \
        attribute_mappings, \
        "Attribute mappings is empty."


    assembled = {}
    for from_, to in attribute_mappings.items():

        try:
            ptr = get_value_from_hierarchy_with_string_reference(record, from_)

        except KeyError:

            # Supports nonexistent parts
            ptr = None


        # If None, we couldn't find this attribute
        if ptr is None:
            ptr = ''

        if ptr.__class__ is list and default_comma_separate_lists is True:
            ptr = ','.join([str(x) for x in ptr])

        assembled[to] = ptr


    return assembled


def pick_from_hierarchical_records_gen(records, attribute_mappings, **kwargs):
    """Given an iterable and a dictionayr of mappings (the keys of which can be
    hierarchical), yield a dictionary of mapped values for each record.
    """

    for record in records:
        assembled = \
            pick_from_hierarchical_record(
                record,
                attribute_mappings,
                **kwargs)

        yield assembled


class BaseTreeNode(riu.entity.BaseEntity):
    @property
    def children(self):
        raise NotImplementedError()


def _build_tree_node(
        record_index, out_node_index, child_attribute, node_id,
        node_factory_fn=None):

    record = record_index[node_id]
    child_ids = record.get(child_attribute, [])


    # Build children

    children = \
        map(
            lambda child_id: \
                _build_tree_node(
                    record_index,
                    out_node_index,
                    child_attribute,
                    child_id,
                    node_factory_fn=node_factory_fn),

            child_ids)

    children = list(children)


    # Construct node

    if node_factory_fn is None:
        node = record.copy()
        node['children'] = children

    else:
        node = node_factory_fn(record, children)

        assert \
            issubclass(node.__class__, BaseTreeNode) is True, \
            "Node factory is expected to return an instance of " \
                "`BaseTreeNode`: [{}] {}".format(node.__class__.__name__, node)


    # Set into node lookup

    out_node_index[node_id] = node


    return node


def tree_build_down(
        records, id_attribute, child_attribute, node_factory_fn=None):
    """Build a tree of `BaseTreeNode` objects. `node_factory_fn` should take a
    record and a list of child nodes and return a `BaseTreeNode`.
    """

    # Index all nodes

    record_index = {
        record[id_attribute]: record
        for record
        in records
    }


    # Find root

    leaf_ids = []
    parent_ids_s = set(record_index.keys())
    for record in record_index.values():
        id_ = record[id_attribute]
        child_ids = record.get(child_attribute, [])

        if not child_ids:
            leaf_ids.append(id_)

        else:
            for child_id in child_ids:
                parent_ids_s.discard(child_id)


    assert \
        len(parent_ids_s) == 1, \
        "There wasn't exactly one root node: {}".format(list(parent_ids_s))

    parent_ids = list(parent_ids_s)
    root_id = parent_ids[0]


    # Build tree

    node_index = {}

    root_node = \
        _build_tree_node(
            record_index,
            node_index,
            child_attribute,
            root_id,
            node_factory_fn=node_factory_fn)


    return node_index, root_id, leaf_ids


def scalar_tree_build_down(
        node, name_extractor_fn, flat_node_factory_fn=None,
        new_index_current_level=None):
    """Transform a tree from complex nodes to scalar nodes. Expects nodes to
    inherit from `BaseTreeNode`. One possible use-case would be to create
    multilevel string-lookup dictionaries, for example:

    string1 -> string2 -> string3 -> [None] -> value
    """

    if new_index_current_level is None:
        new_index_current_level = {}

    if flat_node_factory_fn is None:
        def flat_node_factory_fn(value, children_index):
            children_index[None] = value
            return children_index


    children_index = {}

    for child in node.children:
        scalar_tree_build_down(
            child,
            name_extractor_fn,
            flat_node_factory_fn=flat_node_factory_fn,
            new_index_current_level=children_index)

    new_flat_node = \
        flat_node_factory_fn(
            node.id,
            children_index)

    name = name_extractor_fn(node)

    new_index_current_level[name] = new_flat_node

    return new_index_current_level
